from functools import partial
import importlib
import io
import pkgutil
import sys
from os import isatty
from string import Formatter
from typing import Optional

import click
from requests.compat import urljoin

from goji.auth import get_credentials, set_credentials
from goji.client import JIRAClient
from goji.config import Configuration
from goji.utils import Datetime


def check_login(client) -> None:
    response = client.get('myself', allow_redirects=False)

    if response.status_code == 401:
        raise click.ClickException('Incorrect credentials. Try `goji login`.')


@click.group()
@click.option('--profile', envvar='GOJI_PROFILE', default='default')
@click.option('--base-url', envvar='GOJI_BASE_URL')
@click.option('--email', envvar='GOJI_EMAIL', default=None)
@click.option('--password', envvar='GOJI_PASSWORD', default=None)
@click.pass_context
def cli(
    ctx, profile: str, base_url: str, email: Optional[str], password: Optional[str]
) -> None:
    config = Configuration.load()
    p = config.profiles.get(profile.lower())

    if p:
        if not base_url:
            base_url = p.url

        if p.email and not email:
            email = p.email

    if not p and profile != 'default':
        raise click.ClickException(f'Profile {profile} not found')

    if not base_url:
        raise click.ClickException('JIRA base URL is not configured')

    if not ctx.obj:
        if ctx.invoked_subcommand == 'login':
            ctx.obj = base_url
        elif email and password:
            ctx.obj = JIRAClient(base_url, auth=(email, password))
        elif email and not password:
            raise click.ClickException('Password is not configured.')
        elif not email and password:
            raise click.ClickException('Email is not configured.')
        else:
            email, password = get_credentials(base_url)

            if not email or not password:
                raise click.ClickException(
                    'Authentication not configured. Run `goji login`.'
                )

            ctx.obj = JIRAClient(base_url, auth=(email, password))


@cli.command('whoami')
@click.pass_obj
def whoami(client: JIRAClient) -> None:
    """View information regarding current user"""
    user = client.get_user()
    click.echo(user)


@click.argument('issue_key')
@cli.command('open')
@click.pass_obj
def open_command(client: JIRAClient, issue_key: str) -> None:
    """Open issue in a web browser"""
    url = urljoin(client.base_url, 'browse/%s' % issue_key)
    click.launch(url)


@click.argument('issue_key')
@cli.command()
@click.pass_obj
def show(client: JIRAClient, issue_key: str) -> None:
    """Print issue contents"""
    issue = client.get_issue(issue_key)
    url = urljoin(client.base_url, 'browse/%s' % issue_key)

    click.echo('\x1b[01;32m-> {issue.key}\x1b[0m'.format(issue=issue))
    click.echo('  {issue.summary}\n'.format(issue=issue))

    if issue.description:
        for line in issue.description.splitlines():
            click.echo('  {}'.format(line))

        click.echo('')

    click.echo('  - Status: {issue.status}'.format(issue=issue))
    click.echo('  - Creator: {issue.creator}'.format(issue=issue))
    click.echo('  - Assigned: {issue.assignee}'.format(issue=issue))
    click.echo('  - URL: {url}'.format(url=url))

    if issue.links:
        click.echo('\n  Related issues:')

        for link in issue.links:
            click.echo('    - {}'.format(link))


@click.argument('user', required=False)
@click.argument('issue_key')
@cli.command()
@click.pass_obj
def assign(client: JIRAClient, issue_key: str, user: Optional[str]) -> None:
    """Assign an issue to a user"""

    if user is None:
        user = client.username

    client.assign(issue_key, user)
    click.echo('{} has been assigned to {}.'.format(issue_key, user))


@click.argument('issue_key')
@cli.command()
@click.pass_obj
def unassign(client: JIRAClient, issue_key: str) -> None:
    """Unassign an issue"""

    client.assign(issue_key, None)
    click.echo('{} has been unassigned.'.format(issue_key))


@click.argument('status', required=False)
@click.argument('issue_key')
@cli.command('change-status')
@click.pass_obj
def change_status(client: JIRAClient, issue_key: str, status: Optional[str]) -> None:
    """Change the status of an issue"""
    click.echo('Fetching possible transitions...')
    transitions = client.get_issue_transitions(issue_key)
    if len(transitions) == 0:
        click.echo('No transitions found for {}'.format(issue_key))
        return

    if status is None:
        for index, transition in enumerate(transitions):
            click.echo('{}: {}'.format(index, transition))
        index = click.prompt('Select a transition', type=int)
        if index < 0 or index >= len(transitions):
            click.echo('No transitions match "{}"'.format(index))
            return
    else:
        index = -1
        for idx, transition in enumerate(transitions):
            if transition.name.lower() == status.lower():
                index = idx
        if index < 0:
            click.echo('No transitions match "{}"'.format(status))
            return

    transition = transitions[index]
    client.change_status(issue_key, transition.id)
    click.echo('Okay, the status for {} is now "{}".'.format(issue_key, transition))


@click.argument('issue_key')
@cli.command()
@click.pass_obj
def comments(client: JIRAClient, issue_key: str) -> None:
    comments = client.comments(issue_key)

    for index, comment in enumerate(comments.comments):
        if index != 0:
            print()

        print('=' * 60)
        print(comment.author)
        print(comment)


@click.option('--message', '-m', help='Message to comment.')
@click.argument('issue_key')
@cli.command()
@click.pass_obj
def comment(client: JIRAClient, message: Optional[str], issue_key: str) -> None:
    """Comment on an issue"""

    if not message:
        MARKER = '# Leave a comment on {}'.format(issue_key)
        message = click.edit(MARKER)

    if message is None or len(message) == 0:
        return

    try:
        client.comment(issue_key, message)
        click.echo('Comment created')
    except Exception as e:
        click.echo('Comment:\n')
        click.echo(message)
        click.echo('')
        raise e


@click.argument('issue_key')
@cli.command()
@click.pass_obj
def edit(client: JIRAClient, issue_key: str) -> None:
    """Edit issue description"""
    issue = client.get_issue(issue_key)
    description = click.edit(issue.description)

    assert issue.description
    if description is not None and description.strip() != issue.description.strip():
        try:
            client.edit_issue(issue_key, {'description': description.strip()})
            click.echo(
                'Okay, the description for {} has been updated.'.format(issue_key)
            )
        except Exception as e:
            click.echo('There was an issue saving the new description:')
            click.echo(description)
            click.echo('')
            raise e


@click.argument('attachments', type=click.File('rb'), nargs=-1)
@click.argument('issue_key')
@cli.command()
@click.pass_obj
def attach(client: JIRAClient, issue_key: str, attachments) -> None:
    """Attach file(s) to an issue"""

    if len(attachments) == 0:
        click.echo('No attachments to upload')
        return

    for fp in attachments:
        attachments = client.attach(issue_key, fp)

        for attachment in attachments:
            print(f'Attachment {attachment.filename} uploaded.')


@click.argument('summary')
@click.argument('project', metavar='<key>')
@click.option('--type', '-t', required=True)
@click.option('--component', '-c', multiple=True, help='Adds a component')
@click.option('--label', multiple=True, help='Adds a label')
@click.option('--priority', '-p', help='Sets the issue priority')
@click.option('--description', help='Sets the issue description')
@cli.command()
@click.pass_obj
def create(client, project, summary, type, component, label, priority, description):
    """
    Create a new issue, prompting for description contents

    Example:

    \b
        goji create ENG 'Update installation guide' \\
            --type 'Task' \\
            --component Handbook --component External \\
            --priority Low
    """

    if description is None:
        description = click.edit()

    if description is not None:
        fields = {
            'summary': summary,
            'description': description,
            'project': {'key': project},
        }

        if len(component) > 0:
            fields['components'] = [{'name': x} for x in component]

        if len(label) > 0:
            fields['labels'] = label

        if priority is not None:
            fields['priority'] = {'name': priority}

        if type is not None:
            fields['issuetype'] = {'name': type}

        try:
            issue = client.create_issue(fields)
            click.echo('Issue {} created'.format(issue.key))
        except Exception as e:
            click.echo('Description:\n')
            click.echo(description)
            click.echo('')
            raise e


@cli.command()
@click.pass_obj
def login(base_url: str) -> None:
    """Authenticate with JIRA server"""
    email, password = get_credentials(base_url)
    if email is not None:
        if not click.confirm('This server is already configured. Override?'):
            return

    click.echo('Enter your JIRA credentials')

    email = click.prompt('Email', type=str)
    password = click.prompt('Password', type=str, hide_input=True)

    client = JIRAClient(base_url, auth=(email, password))
    check_login(client)
    set_credentials(base_url, email, password)


@click.argument('type')
@click.argument('outward-issue')
@click.argument('inward-issue')
@click.option('--list-types/--no-list-types', default=False)
@cli.command()
@click.pass_obj
def link(client: JIRAClient, inward_issue: str, outward_issue: str, type: str, list_types: bool) -> None:
    if list_types:
        types = client.get_issue_link_types()
        for link_type in types:
            print(f'{link_type.name} ({link_type.inward} -> {link_type.outward})')
        return

    client.link_issue(inward_issue, outward_issue, type)
    print(f'{inward_issue} linked to {outward_issue}')


@click.argument('query')
@click.option('--limit', type=int)
@click.option('--format', default='{key} {summary}')
@click.option('--count', is_flag=True, help='Return the count of matched issues')
@click.option('--all', is_flag=True, help='Return all pages of issues')
@cli.command()
@click.pass_obj
def search(client: JIRAClient, all: bool, count: bool, format: str, limit: Optional[int], query: str) -> None:
    """Search issues using JQL"""

    if count:
        results = client.search(query, fields=['key'], max_results=1)
        print(results.total)
        return

    formatter = Formatter()
    fields = [v[1] for v in formatter.parse(format) if v[1]]

    if all:
        issues = client.search_all(query, fields=fields)
    else:
        issues = client.search(query, fields=fields, max_results=limit).issues

    for issue in issues:
        format_kwargs = dict(
            key=issue.key,
            summary=issue.summary,
            description=issue.description,
            creator=issue.creator,
            assignee=issue.assignee,
            status=issue.status,
            resolution=issue.resolution,
        )

        if 'key' in fields and not isinstance(sys.stdout, io.TextIOWrapper) and isatty(sys.stdout.fileno()):
            url = urljoin(client.base_url, 'browse/%s' % issue.key)
            format_kwargs['key'] = f'\033]8;;{url}\a{issue.key}\033]8;;\a'

        format_kwargs.update(issue.customfields)

        click.echo(format.replace('\\n', '\n').format(**format_kwargs))


@cli.group('sprint')
def sprint() -> None:
    pass


@sprint.command('create')
@click.argument('board_id', envvar='GOJI_BOARD_ID', type=int)
@click.argument('name')
@click.option('--start', type=Datetime(format='%d/%m/%y'), default=None)
@click.option('--end', type=Datetime(format='%d/%m/%y'), default=None)
@click.pass_obj
def sprint_create(client: JIRAClient, board_id: int, name: str, start, end) -> None:
    """Create a sprint"""

    client.create_sprint(board_id, name, start_date=start, end_date=end)
    click.echo('Sprint created')


# Support for plugins, this API is somewhat unstable, it may change between versions.
def add_plugin_command(plugin, main):
    @cli.command(
        name=getattr(plugin, '__name__'),
        help=getattr(plugin, '__doc__'),
        context_settings=dict(
            ignore_unknown_options=True,
        ),
    )
    @click.pass_obj
    @click.argument('args', nargs=-1, type=click.UNPROCESSED)
    def func(client: JIRAClient, args):
        main(client, args)


plugins = [name for _, name, _ in pkgutil.iter_modules() if name.startswith('goji_')]
for module in plugins:
    try:
        plugin = importlib.import_module(module)
        main = partial(getattr(plugin, '__main__'))
        add_plugin_command(plugin, main)
    except Exception as e:
        print(f'Failed to load plugin {module}: {repr(e)}', file=sys.stderr)
        continue



import re
from typing import List

BOLD_PATTERN = re.compile(r'\*([^\*\n]*)\*')
EMPHASIS_PATTERN = re.compile(r'_([^_]*)_')
COLOR_PATTERN = re.compile(r'{color:([^}]*)}(.*){color}')
LINK_PATTERN = re.compile(r'\[([^|]*)\|([^\]]*)\]')
NOFORMT_PATTERN = re.compile(r'{noformat}(.*){noformat}')
TEXT_EFFECT_PATTERNS = re.compile(
    r'({noformat}.*{noformat}|{color:[^}]*}.*{color}|\*[^\*\n]*\*|_[^_]*_|\[[^|]*\|[^\]]*\])'
)


class Node:
    pass


class Text(Node):
    def __init__(self, value: str):
        self.value = value

    def __repr__(self) -> str:
        return f'<Text {self.value}>'

    def __eq__(self, other) -> bool:
        return isinstance(other, Text) and self.value == other.value


class Bold(Node):
    @classmethod
    def pattern(cls):
        return BOLD_PATTERN

    def __init__(self, value: str):
        self.value = value

    def __repr__(self) -> str:
        return f'<Bold {self.value}>'

    def __eq__(self, other) -> bool:
        return isinstance(other, Bold) and self.value == other.value


class Emphasis(Node):
    @classmethod
    def pattern(cls):
        return EMPHASIS_PATTERN

    def __init__(self, value: str):
        self.value = value

    def __repr__(self) -> str:
        return f'<Emphasis {self.value}>'

    def __eq__(self, other) -> bool:
        return isinstance(other, Emphasis) and self.value == other.value


class Color(Node):
    @classmethod
    def pattern(cls):
        return COLOR_PATTERN

    def __init__(self, color: str, value: str):
        self.color = color
        self.value = value

    def __repr__(self) -> str:
        return f'<Color({self.color}) {self.value}>'

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, Color)
            and self.color == other.color
            and self.value == other.value
        )


class Link(Node):
    @classmethod
    def pattern(cls):
        return LINK_PATTERN

    def __init__(self, url: str, value: str):
        self.url = url
        self.value = value

    def __repr__(self) -> str:
        return f'<Link({self.url}) {self.value}>'

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, Link)
            and self.url == other.url
            and self.value == other.value
        )


def parse_text_effect(value: str) -> Node:
    match = NOFORMT_PATTERN.match(value)
    if match:
        return Text(match.group(1))

    match = Bold.pattern().match(value)
    if match:
        return Bold(match.group(1))

    match = Emphasis.pattern().match(value)
    if match:
        return Emphasis(match.group(1))

    match = Color.pattern().match(value)
    if match:
        return Color(match.group(1), match.group(2))

    match = Link.pattern().match(value)
    if match:
        return Link(match.group(2), match.group(1))

    return Text(value)


def parse(value: str) -> List[Node]:
    """
    https://jira.atlassian.com/secure/WikiRendererHelpAction.jspa?section=all
    """

    return list(
        map(
            parse_text_effect,
            filter(lambda t: len(t) != 0, TEXT_EFFECT_PATTERNS.split(value)),
        )
    )



from datetime import datetime

from click import ParamType


class Datetime(ParamType):
    name = 'date'

    def __init__(self, format: str):
        self.format = format

    def convert(self, value, param, ctx):
        if value is None or isinstance(value, datetime):
            return value

        try:
            return datetime.strptime(value, self.format)
        except ValueError as e:
            self.fail(
                'Could not parse datetime string "{datetime_str}" formatted as {format} ({ex})'.format(
                    datetime_str=value,
                    format=self.format,
                    ex=e,
                ),
                param,
                ctx,
            )



import datetime
import mimetypes
from typing import Any, Dict, Generator, List, Optional

import click
import requests
from requests.auth import AuthBase, HTTPBasicAuth
from requests.compat import urljoin

from goji.models import (Attachment, Comment, Comments, Issue, IssueLinkType,
                         SearchResults, Sprint, Transition, UserDetails)


class JIRAException(click.ClickException):
    def __init__(self, error_messages: List[str], errors):
        self.error_messages = error_messages
        self.errors = errors

    def show(self):
        for error in self.error_messages:
            click.echo(error)

        for (key, error) in self.errors.items():
            click.echo('- {}: {}'.format(key, error))


class NoneAuth(AuthBase):
    """
    Creates a "None" auth type as if actual None is set as auth and a netrc
    credentials are found, python-requests will use them instead.
    """

    def __call__(self, request):
        return request


class JIRAClient(object):
    def __init__(self, base_url: str, auth=None):
        self.session = requests.Session()
        self.base_url = base_url
        self.rest_base_url = urljoin(self.base_url, 'rest/api/2/')

        if auth:
            self.session.auth = HTTPBasicAuth(auth[0], auth[1])
        else:
            self.session.auth = NoneAuth()

    # Methods

    def validate_response(self, response: requests.Response) -> None:
        if response.status_code >= 400 and 'application/json' in response.headers.get(
            'Content-Type', ''
        ):
            error = response.json()
            raise JIRAException(error.get('errorMessages', []), error.get('errors', {}))

    def get(self, path: str, **kwargs) -> requests.Response:
        url = urljoin(self.rest_base_url, path)
        response = self.session.get(url, **kwargs)
        self.validate_response(response)
        return response

    def post(self, path: str, json) -> requests.Response:
        url = urljoin(self.rest_base_url, path)
        response = self.session.post(url, json=json)
        self.validate_response(response)
        return response

    def put(self, path: str, json) -> requests.Response:
        url = urljoin(self.rest_base_url, path)
        response = self.session.put(url, json=json)
        self.validate_response(response)
        return response

    @property
    def username(self) -> Optional[str]:
        if self.session.auth and isinstance(self.session.auth, HTTPBasicAuth):
            return self.session.auth.username

        return None

    def get_user(self) -> Optional[UserDetails]:
        response = self.get('myself', allow_redirects=False)
        response.raise_for_status()
        return UserDetails.from_json(response.json())

    def get_issue(self, issue_key: str) -> Issue:
        response = self.get('issue/%s' % issue_key)
        response.raise_for_status()
        return Issue.from_json(response.json())

    def get_issue_transitions(self, issue_key: str) -> List[Transition]:
        response = self.get(
            'issue/%s/transitions' % issue_key, params={'expand': 'transitions.fields'}
        )
        response.raise_for_status()
        return list(map(Transition.from_json, response.json()['transitions']))

    def change_status(self, issue_key: str, transition_id: str) -> None:
        data = {'transition': {'id': transition_id}}
        self.post('issue/%s/transitions' % issue_key, data)

    def edit_issue(self, issue_key: str, updated_fields) -> None:
        data = {'fields': updated_fields}
        self.put('issue/%s' % issue_key, data)

    def attach(self, issue_key: str, attachment) -> List[Attachment]:
        media_type = mimetypes.guess_type(attachment.name)
        files = {
            'file': (attachment.name, attachment, media_type[0]),
        }
        url = urljoin(self.rest_base_url, f'issue/{issue_key}/attachments')
        response = self.session.post(
            url,
            headers={'X-Atlassian-Token': 'no-check'},
            files=files,
        )
        self.validate_response(response)
        return list(map(Attachment.from_json, response.json()))

    def get_issue_link_types(self) -> List[IssueLinkType]:
        response = self.get('issueLinkType')
        return list(map(IssueLinkType.from_json, response.json()['issueLinkTypes']))

    def link_issue(self, outward_issue: str, inward_issue: str, type: str):
        data = {
            'type': {
                'name': type,
            },
            'inwardIssue': {
                'key': inward_issue,
            },
            'outwardIssue': {
                'key': outward_issue,
            },
        }
        self.post('issueLink', data)

    def create_issue(self, fields) -> Issue:
        response = self.post('issue', {'fields': fields})
        return Issue.from_json(response.json())

    def assign(self, issue_key: str, name: Optional[str]) -> None:
        response = self.put('issue/%s/assignee' % issue_key, {'name': name})
        response.raise_for_status()

    def comment(self, issue_key: str, comment: str) -> Comment:
        response = self.post('issue/%s/comment' % issue_key, {'body': comment})
        return Comment.from_json(response.json())

    def comments(self, issue_key: str) -> Comments:
        response = self.get('issue/%s/comment' % issue_key)
        return Comments.from_json(response.json())

    def search(
        self,
        query: str,
        fields: Optional[List[str]] = None,
        max_results: Optional[int] = None,
        start_at: Optional[int] = None,
    ) -> SearchResults:
        body: Dict[str, Any] = {'jql': query}

        if start_at:
            body['startAt'] = start_at

        if max_results is not None:
            body['maxResults'] = max_results

        if fields:
            body['fields'] = fields

        response = self.post('search', body)
        response.raise_for_status()
        return SearchResults.from_json(response.json())

    def search_all(
        self,
        query: str,
        fields: Optional[List[str]] = None,
    ) -> Generator[Issue, None, None]:
        issues = 0

        while True:
            results = self.search(query, fields, start_at=issues)
            issues += len(results.issues)

            for issue in results.issues:
                yield issue

            if len(results.issues) == 0:
                break

            if issues >= results.total:
                break

    def create_sprint(
        self,
        board_id: int,
        name: str,
        start_date: Optional[datetime.datetime] = None,
        end_date: Optional[datetime.datetime] = None,
    ) -> Sprint:
        payload = {
            'originBoardId': board_id,
            'name': name,
        }

        if start_date:
            payload['startDate'] = start_date.isoformat()

        if end_date:
            payload['endDate'] = end_date.isoformat()

        url = urljoin(self.base_url, 'rest/agile/1.0/sprint')
        response = self.session.post(url, json=payload)
        self.validate_response(response)
        return Sprint.from_json(response.json())



import re
from netrc import netrc
from os import chmod, path
from stat import S_IRUSR, S_IWUSR
from textwrap import dedent
from typing import Optional, Tuple

from requests.compat import urlparse


def get_credentials(base_url: str) -> Tuple[Optional[str], Optional[str]]:
    hostname = urlparse(base_url).hostname
    try:
        hosts = netrc().hosts
        if hostname in hosts:
            return (hosts[hostname][0], hosts[hostname][2])
    except:
        pass

    return (None, None)


def set_credentials(base_url: str, email, password) -> None:
    hostname = urlparse(base_url).hostname
    assert hostname
    filepath = path.expanduser('~/.netrc')
    if path.isfile(filepath):
        rcfile = open(filepath)
        contents = rcfile.read()
        rcfile.close()
        pattern = r'machine {}\n(\s+(login|password).*)+\n?'
        matcher = re.compile(pattern.format(re.escape(hostname)), re.MULTILINE)
        contents = matcher.sub('', contents)
        with open(filepath, 'w') as rcfile:
            rcfile.write(contents)
            rcfile.write(
                dedent(
                    """\n\
                    machine {}
                      login {}
                      password {}"""
                ).format(hostname, email, password)
            )

    else:
        with open(filepath, 'w') as rcfile:
            rcfile.write(
                dedent(
                    """\
                    machine {}
                      login {}
                      password {}"""
                ).format(hostname, email, password)
            )
        chmod(filepath, S_IWUSR | S_IRUSR)



from goji.text_parser import Bold, Color, Emphasis, Link, Text, parse

ANSI_FOREGROUND_COLOR = {
    'red': '31',
    'green': '32',
}


def text_formatting_to_ansi(value: str) -> str:
    result = ''

    for node in parse(value):
        if isinstance(node, Text):
            result += node.value
        elif isinstance(node, Bold):
            result += f'\x1b[1m{node.value}\x1b[0m'
        elif isinstance(node, Emphasis):
            result += f'\x1b[3m{node.value}\x1b[0m'
        elif isinstance(node, Color):
            if node.color in ANSI_FOREGROUND_COLOR:
                result += (
                    f'\x1b[{ANSI_FOREGROUND_COLOR[node.color]}m{node.value}\x1b[0m'
                )
            else:
                result += node.value
        elif isinstance(node, Link):
            result += f'\x1b]8;;{node.url}\a{node.value}\x1b]8;'

    return result
